/*------------------------------------------------------------*- c++ -*-\
|									|
|		      __   __    ____ _____ ____			|
|		      \ \ / /_ _/ ___|_   _|___ \			|
|		       \ V / _` \___ \ | |   __) |			|
|			| | (_| |___) || |  / __/			|
|			|_|\__,_|____/ |_| |_____|			|
|									|
|				core system				|
|							  (C) SuSE GmbH |
\-----------------------------------------------------------------------/

   File:	Y2Namespace.h
		a generic interface for accessing a namespace from YCP interpreter

   Author:	Stanislav Visnovsky <visnov@suse.cz>
   Maintainer:	Stanislav Visnovsky <visnov@suse.cz>

/-*/

#ifndef Y2Namespace_h
#define Y2Namespace_h

#include <string>
using std::string;

#include "ycp/YCPValue.h"
#include "ycp/YCPVoid.h"
#include "ycp/Type.h"

#include "SymbolEntry.h"

class SymbolTable;
class Point;
class Y2Function;
class StaticDeclaration;
class SymbolTable;

/**
 * Y2Namespace implements a hash(?) table of nested(?) SymbolEntries and
 * allows to look them up
 */
class Y2Namespace {
protected:
    typedef vector<SymbolEntryPtr> symbols_t;

    SymbolTable* m_table;
    unsigned int m_symbolcount;
    symbols_t m_symbols;

    friend class SymbolTable;

    // add symbol to namespace, it now belongs here
    // returns the index into m_symbols
    //
    // this is used for blocks with a local environment but no table
    unsigned int addSymbol (SymbolEntryPtr sentry);

    // add symbol _and_ enter into table for lookup
    //
    // this is used for namespaces with a global environment and a table
    void enterSymbol (SymbolEntryPtr sentry, Point *point = 0);

    // lookup symbol by name in m_symbols
    SymbolEntryPtr lookupSymbol (const char *name) const;


    // release symbol from m_symbols
    //   it's no longer owned by this block but by a ysFunction()
    void releaseSymbol (unsigned int position);

    bool m_initialized;

public:

    Y2Namespace ();

    virtual ~Y2Namespace() noexcept(false);

    // end of symbols, finish and clean up m_symbols
    void finish ();

    //! what namespace do we implement
    virtual const string name () const;
    //! used for error reporting
    virtual const string filename () const = 0;

    //! gives the number of symbol declarations
    //  e.g. needed for function declarations which keep their symbolic
    //   parameters in a Y2Namespace
    virtual unsigned int symbolCount () const;

    //! access to definitions of this namespace
    // bytecode uses unsigneds
    virtual SymbolEntryPtr symbolEntry (unsigned int position) const;

    //! unparse. useful for debugging
    virtual string toString () const;

    // just m_symbols, for debugging and YBlock::toString
    string symbolsToString () const;

    //! called when evaluating the import statement
    // constructor is handled separately
    virtual YCPValue evaluate (bool cse = false) = 0;

    //! get our whole symbol table?
    virtual SymbolTable* table () const;

    // this will ensure existence of the table.
    // after calling this function @ref table will always return a valid pointer
    void createTable ();

    /**
     * Creates a function call instance, which can be used to call a
     * function from this namespace. The object is NOT owned anymore by this
     * instance, the caller can (and should) delete it.
     *
     * @param name	name of the required function
     * @param type	the type of the function (needed for overloading)
     * @return 		an object, that can be used to call the function, or NULL on error
     */
    virtual Y2Function* createFunctionCall (const string name, constFunctionTypePtr type) = 0;

    // push all local variables to stack, uses SymbolEntry::push()
    void pushToStack ();

    // pop all local variables from stack, uses SymbolEntry::pop()
    void popFromStack ();

    // ensure that the namespace is initialized
    virtual void initialize ();

};

#define ERROR_NAMESPACE_NAME "_ErrorNamespace"
class Y2ErrorNamespace : public Y2Namespace
{
public:

  /**
   * Constructs a new namespace in case an error happens
   *
   * @param summary short summary what error happened
   * @param details can be longer details for the error e.g. backtrace
   */
  Y2ErrorNamespace(const std::string & summary, const std::string & details);

  ~Y2ErrorNamespace();

  /**
   * short summary what error happened
   */
  const std::string & summary() const;

  /**
   * longer details for error
   */
  const std::string & details() const;

  const string name () const { return ERROR_NAMESPACE_NAME; }

  const string filename () const { return "virtual_error_place"; }

  /**
   * creating a function call for the error namespace always fails
   */
  virtual Y2Function* createFunctionCall (const string name, constFunctionTypePtr type)
  {
    return NULL;
  }

  /**
   * Does nothing for the error namespace
   */
  YCPValue evaluate (bool cse = false);

private:

  string summary_;

  string details_;
};

inline bool isErrorNamespace(const Y2Namespace * inst)
{
  return inst->name() == ERROR_NAMESPACE_NAME;
}

inline Y2ErrorNamespace * toErrorNamespace(Y2Namespace * inst)
{
  return dynamic_cast<Y2ErrorNamespace*>(inst);
}

#endif // Y2Namespace_h
